from dataclasses import dataclass
from typing import Callable


@dataclass
class Progress:
    message: str


ProgressCallback = Callable[[Progress], None]


def print_progress(progress: Progress):
    print(f"{progress.message}...", flush=True)
